// listing2.cpp - the thread object
#include "defines.h"
#include <sys/types.h>
#include <stdio.h>
#include <unistd.h>
#include <mt.h>
#include <sys/signal.h>
#include <thread.h>
#include <sys/lwp.h>
#include "listing2.hpp"

void *ThreadFunc( void * );
void *SigWaiter( void * );

// BlockAllSignals - Every spawned thread calls this to block signals.
void BlockAllSignals()
{
sigset_t SigSet;
if( sigfillset (&SigSet) < 0 )
	printf( "P%d LWP%d - Thread ?\t sigfillset failed\n", getpid(), _lwp_self() );
if( thr_sigsetmask( SIG_SETMASK, &SigSet, NULL ) < 0 )
	printf( "P%d LWP%d - Thread ?\t sigsetmask failed\n", getpid(), _lwp_self() );
}

// SigWaiter - the start function for the SignalHandlerThread.  All it
// does is sigwait on the signal set created and maintained over in
// main() and print out the number of whatever signal it gets.
void *SigWaiter( void *pV )
{
int sig;

Thread *pT = (Thread *)pV;
BlockAllSignals();

while( 1 )
	{
	if(( sig = sigwait( &SigWaitSet )) < 0 )
		printf( "P%d LWP%d - Thread ?\t sigwait failed %d\n", getpid(), _lwp_self(), sig );
	else
		printf( "Signal %d caught by SignalHandlerThread\n", sig );
	if( sig == SIGINT )
		bKeepRunning = FALSE;
	if( pT->Ended )
		{
		printf( "P%d LWP%d - SignalHandlerThread %d\t ended\n", getpid(), _lwp_self(), pT->GetID() );
		break;
		}
	if( pT->bVerbose )
		printf( "P%d LWP%d - SignalHandlerThread %d\t\n", getpid(), _lwp_self(), pT->GetID() );
	}
return( NULL );
}

Thread::Thread( long flags, void *(*start_routine)(void *arg) )
{

unsigned long lFlags = flags; // BOUND, SUSPENDED ...
Ended = FALSE;
bVerbose = TRUE;

iCreateError = thr_create(NULL, 0, start_routine, this, lFlags, &tid );

if( iCreateError != 0 )
	{
	ReportError( iCreateError );
	return;
	}
else
	printf( "P%d d LWP%d - Created thread %d\t (%d)\n", getpid(), _lwp_self(), tid, iCreateError );

sched_param_t sparam;
iRet = thr_getscheduler(tid, &sparam);

//sparam.policy=SCHED_RR;
//iRet = thr_setscheduler(tid, &sparam);
}

// ThreadFunc - this is the start function for all threads in this
// program. These few lines are all that the spawned threads ever
// execute.  The argument pV is a pointer to a Thread object as
// defined in listing2.hpp
void *ThreadFunc( void *pV )
{
Thread *pT = (Thread *)pV;
static unsigned long ulIteration;

BlockAllSignals();
while( 1 )
	{
	sleep( 1 );
	if( pT->Ended )
		{
		printf( "P%d LWP%d - Thread %d\t ended\n", getpid(), _lwp_self(), pT->GetID() );
		break;
		}
	if( pT->bVerbose )
		printf( "P%d LWP%d - Thread %d\t iteration %ld\n", getpid(), _lwp_self(), pT->GetID(), ulIteration );
	ulIteration++;
	}
return( NULL );
}

// Thread constructor - create a thread with the specified flags, and
// ThreadFunc as its start function.  Do a little more elaborate 
// error reporting here as this is where we might run into the 
// MAXULWP roadblock.
Thread::Thread( long flags )
{

unsigned long lFlags = flags; // BOUND, SUSPENDED ...
Ended = FALSE;
bVerbose = TRUE;

iCreateError = thr_create(NULL, 0, ThreadFunc, this, lFlags, &tid );

if( iCreateError != 0 )
	{
	ReportError( iCreateError );
	return;
	}
else
	printf( "P%d d LWP%d - Created thread %d\t (%d)\n", getpid(), _lwp_self(), tid, iCreateError );

sched_param_t sparam;
iRet = thr_getscheduler(tid, &sparam);

//sparam.policy=SCHED_RR;
//iRet = thr_setscheduler(tid, &sparam);
}

// End - Just set the bool Ended.  This will terminate the loop in 
// ThreadFunc, causing a return (and implicit thr_exit).
bool Thread::End()
{
Ended = TRUE;
printf( "P%d LWP%d - Ended thread %d\t (%d)\n", getpid(), _lwp_self(), tid, iRet );
return( FALSE );
}

// Continue - Restarts this suspended thread
bool Thread::Continue()
{
iRet = thr_continue( tid );
printf( "P%d LWP%d - Contined thread %d\t (%d)\n", getpid(), _lwp_self(), tid, iRet );
return( FALSE );
}

// Suspend - Suspends this thread
bool Thread::Suspend()
{
iRet = thr_suspend( tid );
printf( "P%d LWP%d - Suspended thread %d\t (%d)\n", getpid(), _lwp_self(), tid, iRet );
return( FALSE );
}

// Kill - Sends the specified signal to this thread
bool Thread::Kill(int sig)
{
iRet = thr_kill( tid, sig );
printf( "P%d LWP%d - sent sig %d to thread %d\t (%d)\n", getpid(), _lwp_self(), sig, tid, iRet );
if( sig == SIGTERM )
	Ended = TRUE;
return( FALSE );
}


// BoundThread - Create a bound&suspended Thread object
BoundThread::BoundThread() : Thread( THR_BOUND|THR_SUSPENDED )
{
}


// MultiplexedThread - Create a multiplexed&suspended Thread object
MultiplexedThread::MultiplexedThread() : Thread( THR_SUSPENDED )
{
}

// SignalHandlerThread - Create a bound thread object running the
// sigwaiter function.
SignalHandlerThread::SignalHandlerThread() : Thread( THR_BOUND, SigWaiter )
{
}

void Thread::ReportError( int iError )
{

switch( iError )
	{
	case ENOMEM:
		printf( "Insufficient memory to complete thr_create.\n" );
		break;

        case EINVAL:
		printf( "One of the following errors occurred:\n" );
		printf( "\tstack_size is zero and stack_address is not NULL.\n" );
        	printf( "\tstack_size, after shrinking to aligned offsets, is\
                      \tsmaller than an implementation-defined lower bound as\
                      \treturned by thr_minstack.\n" );
        	printf( "\tstart_routine is NULL.\n" );
        	printf( "\tThe process is being traced via ptrace.\n" );
		break;

        case EAGAIN:
		printf( "A resource limit would be exceeded if the call\n\
                      succeeded.  Note that EAGAIN cannot always be detected.\n" );
		break;
	default:
		printf( "Unknown error %d\n", iError );
		break;
	}
}
